(*====================================================================
    File rDBUtils.pas

    Components library for Delphi, Kylix and BCB++.
    ,       .
    Procedures, functions and types for working with database.

    Copyright  1998-2008, PSTG
    http://www.polesoft.ru
    mailto:support@polesoft.ru

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

======================================================================
$Id: rDBUtils.pas,v 1.24 2008/07/30 07:40:50 savchenko Exp $
======================================================================*)
unit rDBUtils;

{$I POLARIS.INC}
{$J+}

interface

uses Classes, rDBConst, DB, DBConsts, SysUtils, rUtils, MaxMin,
     Windows, Registry, rErrorUtils
     {$IFDEF POLARIS_D6}, SqlTimSt, RxStrUtils, Variants{$ELSE}, StrUtils{$ENDIF};

type
  TrDBOption =
    (tboFirst, tboLast, tboEmpty,
     tboCanModify, tboInserting, tboModifying, tboEditing,
     tboAfterEditAssigned, tboNULL);
  TrDBOptionSet = set of TrDBOption;

  TrDBAction = (tbaCustom, tbaFirst, tbaPrior, tbaNext, tbaLast, tbaRefresh,
                tbaInsert, tbaAppend, tbaDelete, tbaEdit, tbaView, tbaPost, tbaCancel,
                tbaConfirmDelete, tbaNULL);

  TrDBEngineType = (etBDE, etDBX, etIBX);

  TrDBDeleteMessage = function (ADataSet: TDataSet; AMulti: Boolean): Boolean;

const
  // 666 - set in initialization section
  DBActionShortCuts: array[TrDBAction] of TShortCut =
    (0, 666, VK_UP, VK_DOWN, 666, 666,
     VK_INSERT, VK_INSERT, VK_DELETE, VK_F4, 666, 666, VK_ESCAPE,
     VK_DELETE, VK_DELETE);

  FindFieldKey = 'FindField';

  Negatives: array[TrDBAction] of TrDBOptionSet =
    ([],[tboFirst, tboEditing],[tboFirst, tboEditing],[tboLast, tboEditing],[tboLast, tboEditing],[tboEditing],
     [tboEditing],[tboEditing],[tboEmpty, tboEditing],[tboEmpty, tboEditing],[tboEmpty, tboEditing],[],[],
     [tboEmpty, tboEditing],[tboNULL]);

  Positives: array[TrDBAction] of TrDBOptionSet =
    ([],[],[],[],[],[],
     [tboCanModify],[tboCanModify],[tboCanModify],[tboCanModify],[tboAfterEditAssigned],[tboEditing],[tboEditing],
     [tboCanModify],[tboEditing]);

  DBActionResources: array[TrDBAction] of string[20] =
    (crEmptyStr,'RDBN_FIRST','RDBN_PRIOR','RDBN_NEXT','RDBN_LAST','RDBN_REFRESH',
     'RDBN_INSERT','RDBN_INSERT','RDBN_DELETE','RDBN_EDIT','RDBN_VIEW','RDBN_POST','RDBN_CANCEL',
     'RDBN_DELETE','NULL_BTN');

  DBActionHints: Array[TrDBAction] of Pointer = (
    @RDBN_Custom,
    @RDBN_First,
    @RDBN_Prior,
    @RDBN_Next,
    @RDBN_Last,
    @RDBN_Refresh,
    @RDBN_Insert,
    @RDBN_Insert,
    @RDBN_Delete,
    @RDBN_Edit,
    @RDBN_View,
    @RDBN_Post,
    @RDBN_Cancel,
    @RDBN_Delete,
    @RDBN_NULL);

  { TrActions }
  RDBA_Caption: Array[TrDBAction] of Pointer = (
    @RDBA_Custom,
    @RDBA_First,
    @RDBA_Prior,
    @RDBA_Next,
    @RDBA_Last,
    @RDBA_Refresh,
    @RDBA_Insert,
    @RDBA_Insert,
    @RDBA_Delete,
    @RDBA_Edit,
    @RDBA_View,
    @RDBA_Post,
    @RDBA_Cancel,
    @RDBA_Delete,
    @RDBA_NULL
  );

  DBImageIndexes: array[TrDBAction] of Integer = (-1,0,1,2,3,4,5,5,6,7,8,9,10,6,11);

var
  GetRecordNameProc: procedure (const ADataSet: TDataSet; AQ: Integer; var ARecordName: string);
  GetTableNameProc: procedure (const ADataSet: TDataSet; var ARecordName: string);
  DBDeleteMessage: TrDBDeleteMessage;  

//    dataset  field
// Check options on dataset and field
function TestOptions(DataSet: TDataSet; Field: TField; PositiveOptions,NegativeOptions: TrDBOptionSet): Boolean;

//   
// Recalculate calculating fields
procedure ReCalcField(DataSet: TDataSet);

//       
//      
// Format "Name" field: cutting blanks and convert first letter to upper case
procedure FormatNameField(Field: TStringField);

//   
// Format "Initials" field
procedure FormatFIOField(Field: TStringField);

//    DataSet
// Change dataset filter
procedure ChangeFilter(DataSet: TDataSet; NewFilter: string);

//     
// Set filter value from system registry
procedure SetDBFilter(AModule: TComponent; Reg: TRegistry; ARegKey: String);

{$IFDEF POLARIS_D4}
//   
// Assign parameter (TParam) value
procedure AssignParam(Param: TParam; Value: Variant);

//    
// Clear all TParam's values
procedure ClearParams(Params: TParams);

//     
// Assign parameter (TParam) value from field
procedure AssignParamFromField(Param: TParam; Field: TField);


//   
// Assign field (TField) value
procedure AssignField(Field: TField; Value: Variant);
{$ENDIF}

//   
// DB errors handler
procedure DBErrorEHProc(E: Exception; Handler: TrError);

//    
// DB server errors handler
procedure DBSrvErrorEHProc(E: Exception; Handler: TrError);

//    dataset  field
// Execute action with dataset or field
function DoAction(DataSet: TDataSet; Field: TField; Action: TrDBAction): Boolean;

//   
// Connect params conversion
procedure EngineParamsConvert(aFrom, aTo: TrDBEngineType; aParamsFrom, aParamsTo: TStrings);

//  
// Copy record
procedure CopyRecord(Source, Dest: TDataSet; ByName: Boolean;
  ReadOnlyIgnore: Boolean {$IFDEF POLARIS_D4}= False{$ENDIF});

//  
// Copy structure
procedure CopyStructure(aSource, aDest: TDataSet);

//  
//conditional dataset refresh
procedure RefreshDataSet(DataSet: TDataSet; KeyField: String = crEmptyStr; UseBookMarks: Boolean = false);

implementation

type
  TCrackDataSet=class(TDataSet);

function TestOptions(DataSet: TDataSet; Field: TField; PositiveOptions,NegativeOptions: TrDBOptionSet): Boolean;
var
  I: TrDBOption;

  function TestOption(O: TrDBOption): Boolean;
  begin
    with DataSet do
    case O of
      tboFirst     : Result := BOF;
      tboLast      : Result := EOF;
      tboEmpty     : Result := (not Active) or (Eof and Bof);
      tboCanModify : Result :=
        (not Assigned(Field) and CanModify) or (Assigned(Field) and Field.CanModify);
      tboInserting : Result := State=dsInsert;
      tboModifying : Result := State=dsEdit;
      tboEditing   : Result := State in [dsInsert,dsEdit];
      tboAfterEditAssigned : Result := Assigned(AfterEdit);
      tboNULL      : Result := Assigned(Field) and Field.IsNull;
    else
      Result := False;
    end;
  end;

begin
  Result := Assigned(DataSet) and DataSet.Active;
  if (PositiveOptions=[]) and (NegativeOptions=[]) then exit;
  if Result and (PositiveOptions<>[]) then
    for I := Low(TrDBOption) to High(TrDBOption) do
      if I in PositiveOptions then begin
        Result := Result and TestOption(I);
        if not Result then Break;
      end;
  if Result and (NegativeOptions<>[]) then
    for I := Low(TrDBOption) to High(TrDBOption) do
      if I in NegativeOptions then begin
        Result := Result and not TestOption(I);
        if not Result then Break;
      end;
end;

procedure ReCalcField(DataSet: TDataSet);
begin
  if DataSet <> nil then
  TCrackDataSet(DataSet).GetCalcFields(DataSet.ActiveBuffer);
end;

procedure FormatNameField;
begin
  with Field do begin
    if not(DataSet.State in [dsEdit,dsInsert]) then exit;
    if Length(Trim(AsString))=0 then begin
      AsVariant := Null;
      exit;
    end;
    AsString := Trim(AsString);
    AsString := Stuff(AsString,1,1,AnsiUpperCase(String(AsString[1])))
  end;
end;

procedure FormatFIOField;
begin
  with Field do begin
    if not(DataSet.State in [dsEdit,dsInsert]) then exit;
    if Length(Trim(AsString))=0 then begin
      AsVariant := Null;
      exit;
    end;
    FormatNameField(Field);
    AsString := AnsiProperCase(Trim(AsString),[' ','.',',','-']);
  end;
end;

procedure ChangeFilter;
begin
  with DataSet do
    if Filter <> NewFilter then begin
      Filter := NewFilter;
      Filtered := (Trim(Filter) <> crEmptyStr) or Assigned(OnFilterRecord);
    end;
end;

procedure SetDBFilter(AModule: TComponent; Reg: TRegistry; ARegKey: String);
Var
  i: Integer;
  RealKey: String;
  r, old_r: HKEY;
begin
  For i:=0 to AModule.ComponentCount-1 do
    if (AModule.Components[i] is TDataSet) then begin
      SeparateRegKey(ARegKey + srFilterKey + (AModule.Components[i] as TDataSet).Name, r, RealKey);
      With Reg do begin
        old_r := RootKey;
        try
          RootKey := r;
          if OpenKey(RealKey, False) then
            try
              (AModule.Components[i] as TDataSet).Filter := ReadString(srFilterValue);
            finally
              CloseKey;
            end;
        finally
          RootKey := old_r;
        end;
      end;
    end;
end;

{$IFDEF POLARIS_D4}
procedure AssignParam(Param: TParam; Value: Variant);
begin
  if VarIsNull(Value) then begin
    Param.Clear;
    Param.Bound := True;
  end
  else
{$IFDEF POLARIS_D6}
    if Param.DataType = ftTimeStamp
    then Param.AsSQLTimeStamp := DateTimeToSQLTimeStamp(Value)
    else
{$ENDIF}
    Param.Value := Value;
end;

procedure ClearParams(Params: TParams);
var
  I: Integer;
begin
  for I := 0 to Params.Count - 1 do Params[I].Clear;
end;

procedure AssignParamFromField(Param: TParam; Field: TField);
begin
  if Field.IsNull then begin
    Param.Clear;
    Param.Bound := True;
  end
  else Param.Value := Field.Value;
end;

procedure AssignField(Field: TField; Value: Variant);
begin
  if Assigned(Field) then
    if VarIsEmpty(Value) or VarIsNUll(Value)
    then Field.Clear
    else Field.Value := Value;
end;
{$ENDIF}

procedure DBErrorEHProc(E: Exception; Handler: TrError);
begin
  with Handler do begin
    PreMessage := srDBErrorTitle;
    Message := ConvertMessage(E.Message);
    if Message = crEmptyStr then
      Message := E.Message;
  end;
end;

procedure DBSrvErrorEHProc(E: Exception; Handler: TrError);
begin
  with Handler do begin
    Message := ConvertMessage(E.Message);
    if Message = crEmptyStr then
      Message := E.Message
    else begin
      PreMessage := srDBErrorTitle;
      Description.Text := E.Message;
    end;
  end;
end;

function DoAction(DataSet: TDataSet; Field: TField; Action: TrDBAction): Boolean;
var
  i: Integer;
  KeyField: String;
begin
  if not Assigned(DataSet) then begin
    Result := False;
    exit;
  end;
  Result := True;
  with DataSet do
    case Action of
      tbaFirst  : First;
      tbaPrior  : Prior;
      tbaNext   : Next;
      tbaLast   : Last;
      tbaRefresh: begin
                    if Assigned(Field) then
                      if (Field.FieldKind = fkLookup) and Field.LookupCache then
                        Field.RefreshLookupList
                      else  
                    else
                      for i:=0 to FieldCount-1 do
                        if (Fields[i].FieldKind = fkLookup) and Fields[i].LookupCache then
                          Fields[i].RefreshLookupList;
                    try
                      Refresh
                    except
                      if Assigned(Field) then KeyField := Field.FieldName;
                      RefreshDataSet(DataSet, KeyField, True);
                    end;
                  end;
      tbaInsert : Insert;
      tbaAppend : Append;
      tbaDelete : Delete;
      tbaEdit   : Edit;
      tbaView   : if Assigned(AfterEdit) then AfterEdit(DataSet)
                  else Result := False;
      tbaPost   : Post;
      tbaCancel : Cancel;
      tbaConfirmDelete: if Assigned(DBDeleteMessage) and DBDeleteMessage(DataSet, False) then Delete else Result := False;
      tbaNULL   : if Field <> nil then Field.Clear else Result := False;
    else
      Result := False;
    end;
end;

procedure EngineParamsConvert(
  aFrom, aTo: TrDBEngineType;
  aParamsFrom, aParamsTo: TStrings);

  type
    TParamType = (ptHst, ptDbs, ptUsr, ptPwd, ptRol, ptLng);
    TParamNames = array [Low(TParamType)..High(TParamType), Low(TrDBEngineType)..High(TrDBEngineType)] of string;
    TDriverType = (dtIB, dtOracle, dtInformix, dtMSSQL);
    TDriverNames = array [Low(TDriverType)..High(TDriverType), etBDE..etDBX] of string;

  const
    cParamNames: TParamNames = (
    {--------etBDE-----------etDBX------------etIBX----------}
    {ptHst} ('SERVER NAME',  'HostName',      ''             ),
    {ptDbs} ('DATABASE NAME','Database',      'DatabaseName' ),
    {ptUsr} ('USER NAME',    'User_Name',     'user_name'    ),
    {ptPwd} ('PASSWORD',     'Password',      'password'     ),
    {ptRol} ('ROLE NAME',    'RoleName',      'sql_role_name'),
    {ptLng} ('LANGDRIVER',   'ServerCharSet', 'lc_ctype'     )
    );

    cDriverNames: TDriverNames = (
    {-------------etBDE-------etDBX------}
    {dtIB}       ('INTRBASE', 'Interbase'),
    {dtOracle}   ('ORACLE',   'Oracle'   ),
    {dtInformix} ('INFORMIX', 'Informix' ),
    {dtMSSQL}    ('MSSQL',    'MSSQL'    )
    );

  cDriverName = 'DriverName';

  function GetDriverType(aDriverName: string; aEngineType: TrDBEngineType): TDriverType;
  var
    t: TDriverType;
  begin
    Result := dtIB;
    if aDriverName = crEmptyStr then Exit;
    for t := Low(TDriverType) to High(TDriverType) do
      if aDriverName = cDriverNames[t, aEngineType] then begin
        Result := t;
        Break;
      end;
  end;

  function ConvBDELang(L: string; toBDE: Boolean): string;
  const
    cCountLangs = 6;
  type
    TLangNames = array [1..cCountLangs, 1..2] of string;
  const
    cNames: TLangNames = (
      ('DBWINUS0', 'WIN1250'),
      ('ascii',    'WIN1250'),
      ('ancyrr',   'WIN1251'),
      ('ancyrr',   'CL8MSWIN1251'),
      ('cyrr',     'DOS866'),
      ('db866ru0', 'DOS866')
    );
  var
    i: Integer;
  begin
    for i := 1 to cCountLangs do
      if (i = 1) or
        (toBDE and (cNames[i, 2] = L)) or
        ((not toBDE) and (cNames[i, 1] = L))
      then begin
        if toBDE then Result := cNames[i, 1] else Result := cNames[i, 2];
        if i > 1 then Exit;
      end;
  end;

var
  t: TParamType;
  d: TDriverType;
begin
  if aTo = aFrom then aParamsTo.Assign(aParamsFrom)
  else begin
    // driver name
    if aFrom = etIBX
      then d := dtIB
      else d := GetDriverType(aParamsFrom.Values[cDriverName], aFrom);
    if aTo <> etIBX
      then aParamsTo.Values[cDriverName] := cDriverNames[d, aTo];
    // params
    for t := Low(TParamType) to High(TParamType) do begin
      // for IB and oracle [SERVER NAME] -> [Database]
      if (etBDE in [aFrom, aTo]) and (d in [dtIB, dtOracle]) and (t in [ptHst, ptDbs])
      then begin
        if (aFrom = etBDE) and (t = ptHst)
          then aParamsTo.Values[cParamNames[ptDbs, aTo]] := aParamsFrom.Values[cParamNames[t, aFrom]];
        if (aTo = etBDE) and (t = ptDbs)
          then aParamsTo.Values[cParamNames[ptHst, aTo]] := aParamsFrom.Values[cParamNames[t, aFrom]];
      end
      else
        aParamsTo.Values[cParamNames[t, aTo]] := aParamsFrom.Values[cParamNames[t, aFrom]];
    end;
    // locale
    if etBDE in [aFrom, aTo] then
      if aParamsFrom.Values[cParamNames[ptLng, aFrom]] <> crEmptyStr then
      aParamsTo.Values[cParamNames[ptLng, aTo]] :=
        ConvBDELang(aParamsFrom.Values[cParamNames[ptLng, aFrom]], aTo = etBDE);
  end;
end;

procedure CopyRecord(Source, Dest: TDataSet; ByName: Boolean;
  ReadOnlyIgnore: Boolean);
var I: Integer;
    F: TField;
    ReverReadOnly: Boolean;
begin
  if not (Dest.State in dsEditModes) then DatabaseError(SNotEditing);
  if ByName then begin
    for I := 0 to Source.FieldCount - 1 do begin
      F := Dest.FindField(Source.Fields[I].FieldName);
      if (F <> nil) and (F.FieldKind = fkData) then begin
        if ReadOnlyIgnore then begin
          ReverReadOnly := F.ReadOnly;
          F.ReadOnly    := False;
          F.Value       := Source.Fields[I].Value;
          F.ReadOnly    := ReverReadOnly;
        end else
        if not F.ReadOnly then begin
          F.Value := Source.Fields[I].Value;
        end;
      end;
    end;
  end else begin
    for I := 0 to Min(Source.FieldCount - 1, Dest.FieldCount - 1) do begin
      F := Dest.Fields[i];
      if (F <> nil) then begin
        if ReadOnlyIgnore then begin
          ReverReadOnly := F.ReadOnly;
          F.ReadOnly    := False;
          F.Value       := Source.Fields[I].Value;
          F.ReadOnly    := ReverReadOnly;
        end else
        if not F.ReadOnly then begin
          F.Value := Source.Fields[I].Value;
        end;
      end;
    end;
  end;
end;

procedure CopyStructure(aSource, aDest: TDataSet);
var
  i, Counter: Integer;
begin
  if aDest.Active then aDest.Close;
  for i := aDest.FieldCount - 1 downto 0 do aDest.Fields[I].Free;
  if (aSource = nil) then Exit;
  aSource.FieldDefs.Update;
  if aSource.Fields.Count <> aSource.FieldDefs.Count then begin
    Counter := 0;
    for i := 0 to aSource.FieldDefs.Count - 1 do
      if Assigned(aSource.FindField(aSource.FieldDefs[i].Name)) then
        begin
          Inc(Counter, 1);
          with aDest.FieldDefs.AddFieldDef do
          begin
            Name := aSource.FieldDefs[i].Name;
            Attributes := aSource.FieldDefs[i].Attributes;
            ChildDefs := aSource.FieldDefs[i].ChildDefs;
            DataType := aSource.FieldDefs[i].DataType;
            FieldNo := Counter; {aSource.FieldDefs[i].FieldNo};
            InternalCalcField := aSource.FieldDefs[i].InternalCalcField;
            Precision := aSource.FieldDefs[i].Precision;
            Required := aSource.FieldDefs[i].Required;
            Size := aSource.FieldDefs[i].Size;
          end;
        end;
  end
  else
    aDest.FieldDefs := aSource.FieldDefs;
  for i := 0 to aDest.FieldDefs.Count - 1 do
      aDest.FieldDefs.Items[i].CreateField(aDest);
end;

procedure RefreshDataSet(DataSet: TDataSet; KeyField: String = crEmptyStr; UseBookMarks: Boolean = false);
var
  ID: Variant;
  F: TField;
  BookMk: TBookmark;

  function SetToBookmark(ADataSet: TDataSet; ABookmark: TBookmark): Boolean;
  begin
    Result := False;
    with ADataSet do
      if Active and (ABookmark <> nil) and not IsEmpty and
        BookmarkValid(ABookmark) then
      try
        ADataSet.GotoBookmark(ABookmark);
        Result := True;
      except
      end;
  end;

begin
  with DataSet do begin
    DisableControls;
    BookMk := nil;
    try
      F := nil;
      if (KeyField <> crEmptyStr) then F := DataSet.FindField(KeyField);
      if Assigned(F) then ID := F.Value
                     else if UseBookmarks then
                            if Active then BookMk := GetBookmark;
      Close;
      Open;
      if Assigned(F) then Locate(KeyField, ID, [])
                     else if UseBookMarks then
                            SetToBookmark(DataSet, BookMk);
    finally
      EnableControls;
      if Assigned(BookMk) then FreeBookmark(BookMk);
    end;
  end;
end;

initialization
  AddError('EDatabaseError',DBErrorEHProc);
  AddError('EUpdateError',DBSrvErrorEHProc);

  AddOrReplaceError('EOLEException',DBSrvErrorEHProc);
  {$IFDEF VER190}
  RHandles.Add(Pointer(GetModuleHandle('rxdb11.bpl')));
{$ELSE}
{$IFDEF VER180}
  RHandles.Add(Pointer(GetModuleHandle('rxdb10.bpl')));
{$ELSE}
{$IFDEF VER170}
  RHandles.Add(Pointer(GetModuleHandle('rxdb9.bpl')));
{$ELSE}
{$IFDEF VER150}
  RHandles.Add(Pointer(GetModuleHandle('rxdb7.bpl')));
{$ELSE}
{$IFDEF VER140}
  RHandles.Add(Pointer(GetModuleHandle('rxdb6.bpl')));
{$ELSE}
{$IFDEF VER130}
  RHandles.Add(Pointer(GetModuleHandle('rxdb5.bpl')));
{$ELSE}
{$IFDEF POLARIS_D4}
  RHandles.Add(Pointer(GetModuleHandle('vcldb40.bpl')));
  RHandles.Add(Pointer(GetModuleHandle('rxdb4.bpl')));
{$ELSE}
  RHandles.Add(Pointer(GetModuleHandle('rxdb.dpl')));
{$ENDIF}{$ENDIF}{$ENDIF}{$ENDIF}{$ENDIF}{$ENDIF}{$ENDIF}
end.

