(*====================================================================
    File rDBXUtils.pas

    Components library for Delphi, Kylix and BCB++.
    Procedures, functions and types for working with database with DBX.

    Copyright  2004-2008, PSTG
    http://www.polesoft.ru
    mailto:support@polesoft.ru

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

======================================================================
$Id: rDBXUtils.pas,v 1.6 2008/09/05 09:00:45 savchenko Exp $
======================================================================*)
unit rDBXUtils;

interface

uses
  SysUtils, Classes, IniFiles, SqlExpr, SqlConst, DBXpress, rConst, rDBConst,
  rErrorUtils, rUtils, StrUtils;

function DBXSysDate(ADatabase: TSQLConnection; WithTime: Boolean): TDateTime;
function UpdateDBXDriverParams(aConnection: TSQLConnection): Boolean;
procedure ExecuteSQL(ADatabase: TSQLConnection; SQL: string);
procedure ExecStoredProc(StoredProc: TSQLStoredProc; DB: TSQLConnection);
function FormatDBCode(ADataBase: TSQLConnection; AText, AUserName: String;
  APrefix: Char = '%'): String;

implementation

function DBXSysDate(ADatabase: TSQLConnection; WithTime: Boolean): TDateTime;
type
  TDriverType = (dtOther, dtIB, dtOra);
var
  dt: TDriverType;
  s: string;
begin
  s := UpperCase(ADatabase.DriverName);
  if s = 'INTERBASE' then dt := dtIB  else
  if s = 'ORACLE'    then dt := dtOra else
                          dt := dtOther;
  if dt = dtOther then begin
    if WithTime then Result := SysUtils.Now
    else Result := SysUtils.Date;
  end
  else
    with TSQLQuery.Create(nil) do
    try
      SQLConnection := ADatabase;
      if dt = dtIB
        then SQL.Text := srSQLIBSysDate
        else SQL.Text := srSQLOraSysDate;
      try
        Open;
        Result := Fields[0].AsDateTime;
        if not WithTime then
          Result := Trunc(Result);
        Close;
      except
        if WithTime then Result := SysUtils.Now
        else Result := SysUtils.Date;
      end;
    finally
      Free;
    end
end;

// from SqlExpr.pas:
function GetProfileString(Section, Setting, IniFileName: string): string;
var
  IniFile: TMemIniFile;
  List: TStrings;
begin
  List := TStringList.Create;
  try
    IniFile := TMemIniFile.Create(IniFileName);
    IniFile.ReadSectionValues(Section, List);
    try
      Result := List.Values[ Setting ];
    finally
      IniFile.Free;
    end;
  finally
    List.Free;
  end;
end;
//

type
  TSQLConnection_cr = class(TSQLConnection);
function UpdateDBXDriverParams(aConnection: TSQLConnection): Boolean;
var
  s, vNam, vRes: string;
  slOld: TStringList;
  i: Integer;
begin
  Result := False;
  slOld := TStringList.Create();
  try
    with TSQLConnection_cr(aConnection) do begin
      slOld.Assign(Params);
      vRes := slOld.Values[srpResourceName];
      if DriverName = '' then begin
        if (ConnectionName = '') and (vRes <> '') then ConnectionName := vRes;
        if ConnectionName = ''
          then DriverName := slOld.Values[DRIVERNAME_KEY]
          else DriverName := GetProfileString(ConnectionName, DRIVERNAME_KEY, ConnectionRegistryFile);
      end;
      if DriverName = '' then Exit;
      LibraryName := GetProfileString(DriverName, DLLLIB_KEY, GetDriverRegistryFile(csDesigning in ComponentState));
      if LibraryName = '' then Exit;
      VendorLib := Trim(GetProfileString(DriverName, VENDORLIB_KEY, GetDriverRegistryFile));
      if VendorLib = '' then Exit;
      GetDriverFunc := GetProfileString(DriverName, GETDRIVERFUNC_KEY, GetDriverRegistryFile);
      if GetDriverFunc = '' then Exit;
      if vRes = '' then
        Params.Assign(slOld)
      else begin
        ParamsLoaded := False;
        LoadParamsFromIniFile();
        for i := 0 to slOld.Count - 1 do begin
          vNam := slOld.Names[i];
          if (vNam <> srpResourceName) and (vNam <> DRIVERNAME_KEY) then begin
            s := slOld.Values[vNam];
            if s <> '' then Params.Values[vNam] := s;
          end;
        end;
      end;
    end;
  finally
    slOld.Free();
  end;
  Result := True;
end;

procedure ExecuteSQL(ADatabase: TSQLConnection; SQL: string);
begin
  ADatabase.ExecuteDirect(SQL);
end;

procedure ExecStoredProc(StoredProc: TSQLStoredProc; DB: TSQLConnection);
var
  b: Boolean;
  TD: TTransactionDesc;
begin
  TD.TransactionID := 1;
  TD.IsolationLevel := xilREADCOMMITTED;
  with StoredProc, DB do begin
    b := InTransaction;
    if not b then StartTransaction(TD);
    try
      ExecProc;
      if not b then Commit(TD);
    except
      if not b then Rollback(TD);
      raise;
    end;
  end;
end;

procedure DBErrorEHProc(E: Exception; Handler: TrError);
var
  i: Integer;
begin
  with Handler do begin
    PreMessage := srDBErrorTitle;
    i := Pos('%s', srSQLServerError) - 1;
    if (Copy(srSQLServerError, 1, i) = Copy(E.Message, 1, i)) then
      Message := ConvertMessage(Copy(E.Message, i + 1, Length(E.Message)))
    else
      Message := ConvertMessage(E.Message);
    if Message = '' then Message := E.Message
    else Description.Text := E.Message;
  end;
end;

function FormatDBCode(ADataBase: TSQLConnection; AText, AUserName: String;
  APrefix: Char = '%'): String;
var
  SysDate: TDateTime;
  SValue: String;
  i, n, j: Integer;
  paramList: TStringList;
const
  MaxFuncs = 11;
  TFuncs: array[0..MaxFuncs] of String =
    ('yyyy', 'yy', 'mm', 'dd', 'hh', 'nn', 'ss', 'compname', 'dbusername',
     'dbuser', 'dbname', 'osuser');
begin
  Result := AText;
  SysDate := DBXSysDate(ADataBase, True);
  for i := 0 to MaxFuncs do begin
    while True do begin
      n := Pos(APrefix + AnsiLowerCase(TFuncs[i]), AnsiLowerCase(Result));
      if n = 0 then Break;
      case i of
        0..6: SValue := FormatDateTime(TFuncs[i], SysDate);
        7: SValue := GetCompName;
        8: SValue := AUserName;
        9: SValue := ADataBase.Params.Values['user_name'];
       10: begin
             paramList := TStringList.Create;
             try
               SValue := ADataBase.Params.Values['DataBase'];
               for j := Length(SValue) downto 0 do
                 if (SValue[j] = '/') or (SValue[j] = '\')
               then begin
                 SValue := Copy(SValue, j + 1, MaxInt);
                 Break;
               end;
               SValue := ChangeFileExt(SValue, '');
             finally
               paramList.Free;
             end;
           end;
       11: SValue := GetOSUser;
      end;
      Result := StuffString(Result, n, Length(TFuncs[i]) + 1, SValue);
    end;
  end;
end;

initialization
  AddOrReplaceError('EDatabaseError',DBErrorEHProc);
end.

