(*====================================================================
    File rErrorUtils.pas

    Components library for Delphi, Kylix and BCB++.
    ,    .
    Errors handling classes and utils.

    Copyright  2005-2008, PSTG
    http://www.polesoft.ru
    mailto:support@polesoft.ru

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

======================================================================
$Id: rErrorUtils.pas,v 1.6 2008/08/22 06:16:56 ariafan Exp $
======================================================================*)
unit rErrorUtils;

{$I POLARIS.INC}

interface

uses
  SysUtils, Windows, Classes, IniFiles, rUtils, rConst,
  {$IFDEF POLARIS_D6}RxStrUtils{$ELSE}StrUtils{$ENDIF};

type
  TrError = class(TPersistent)
  private
    FErrorBegins: TStrings;
    FConsts:      TStrings;
    FIniFileName: string;
    FIniLocation: TrFileLocation;
    FLogFile:     string;
    FLogLocation: TrFileLocation;
    FUserName:    string;
    FPreMessage:  string;
    FMessage:     string;
    FDescription: TStrings;
    FSaveLog:     Boolean;
    procedure SetIniFile(Value: string);
    procedure SetIniLocation(Value: TrFileLocation);
    procedure SetMessage(Value: string);
    procedure SetDescription(Value: TStrings);
    function GetConstant(const aName: string): string;
  protected
    { Protected declarations }
  public
    FIniFile: TIniFile;
    constructor Create;
    destructor Destroy; override;
    function  ConvertMessage(mes: string): string;
    procedure AnalyzeException(E: Exception);
    procedure AppException(Sender: TObject; E: Exception);
    procedure SavingLog(BeforeMessage: string {$IFDEF POLARIS_D4}=''{$ENDIF});
    property Constant[const aName: string]: string read GetConstant;
    property IniFile: string       read FIniFileName write SetIniFile;
    property IniLocation: TrFileLocation read FIniLocation write SetIniLocation;
    property LogFile: string       read FLogFile     write FLogFile;
    property LogLocation: TrFileLocation read FLogLocation write FLogLocation;
    property UserName: string      read FUserName    write FUserName;
    property PreMessage: string    read FPreMessage  write FPreMessage;
    property Message: string       read FMessage     write SetMessage;
    property Description: TStrings read FDescription write SetDescription;
    property SaveLog: Boolean      read FSaveLog     write FSaveLog;
  end;

  TrEHProc = procedure(E: Exception; Handler: TrError);
  TrEH = record
    ErrorName: string;
    Proc: TrEHProc;
  end;
  TrpEH = ^TrEH;

  EAnyError = class(Exception)
    FPreMessage: string;
    FDescription: string;
  public
    constructor CreateAny(PreMsg, Msg, Desc: string);
    property PreMessage: string read FPreMessage;
    property Description: string read FDescription;
  end;

procedure AddError(Name: string; Proc: TrEHProc);
procedure AddOrReplaceError(Name: string; Proc: TrEHProc);
procedure DeleteError(Name: string);

procedure AnyErrorEHProc(E: Exception; Handler: TrError);
procedure OSErrorEHProc(E: Exception; Handler: TrError);
procedure ConvertErrorEHProc(E: Exception; Handler: TrError);

var
  RErrors: TList;

implementation

procedure AddError(Name: string; Proc: TrEHProc);
var
  ARecord: TrpEH;
begin
  New(ARecord);
  ARecord^.ErrorName := Name;
  ARecord^.Proc := Proc;
  RErrors.Add(ARecord);
end;

procedure AddOrReplaceError(Name: string; Proc: TrEHProc);
var
  i: Integer;
begin
  for i := 0 to RErrors.Count - 1 do
    if CompareText(TrpEH(RErrors.Items[i])^.ErrorName, Name)=0 then begin
      TrpEH(RErrors.Items[i])^.ErrorName := Name;
      TrpEH(RErrors.Items[i])^.Proc := Proc;
      exit;
    end;
  AddError(Name, Proc);
end;

procedure DeleteError(Name: string);
var
  i: Integer;
begin
  for i := 0 to RErrors.Count - 1 do
    if CompareText(TrpEH(RErrors.Items[i])^.ErrorName, Name)=0 then begin
      Dispose(RErrors.Items[i]);
      RErrors.Delete(i);
      exit;
    end;
end;

procedure OSErrorEHProc(E: Exception; Handler: TrError);
begin
  with Handler do begin
    PreMessage := srOSErrorTitle;
    Message := E.Message;
{$IFDEF POLARIS_D6}
    with EOSError(E) do
{$ELSE}
    with EWin32Error(E) do
{$ENDIF}
      Description.Text := Format(ConvertCodes(srOSErrorDesc),
                                 [ErrorCode, SysErrorMessage(ErrorCode)]);
    SaveLog := Description.Text <> '';
  end;
end;

procedure InOutErrorEHProc(E: Exception; Handler: TrError);
begin
  with Handler do begin
    PreMessage := srIOErrorTitle;
    with EInOutError(E) do begin
      Handler.Message := SysErrorMessage(ErrorCode);
      Description.Text := Format(ConvertCodes(srIOErrorDesc),
                                 [ErrorCode, Handler.Message]);
    end;
    SaveLog := Description.Text <> '';
  end;
end;

procedure ConvertErrorEHProc(E: Exception; Handler: TrError);
begin
  with Handler do begin
    if E is ERangeError then
      PreMessage := srRangeErrorTitle
    else
      PreMessage := srConvErrorTitle;
    Message := Format(ConvertCodes(srConvErrorMes), [E.Message]);
    SaveLog := False;
  end;
end;

{ EAnyError }

procedure AnyErrorEHProc(E: Exception; Handler: TrError);
begin
  with Handler do begin
    PreMessage := ConvertCodes(EAnyError(E).PreMessage);
    Message := ConvertCodes(E.Message);
    Description.Text := ConvertCodes(EAnyError(E).Description);
    SaveLog := Description.Text <> '';
  end;
end;

constructor EAnyError.CreateAny;
begin
  inherited Create(Msg);
  FPreMessage := PreMsg;
  FDescription := Desc;
end;

{ TrError }

constructor TrError.Create;
begin
  inherited;
  FDescription := TStringList.Create;
  FErrorBegins := TStringList.Create;
  FConsts := TStringList.Create;
  FIniFile := nil;
end;

destructor TrError.Destroy;
begin
  inherited;
  FDescription.Free;
  if Assigned(FIniFile) then FIniFile.Free;
  FErrorBegins.Free;
  FConsts.Free;
end;

procedure TrError.SetIniFile(Value: string);
begin
  if Assigned(FIniFile) then begin
    FIniFile.Free;
    FIniFile := nil;
  end;
  FIniFileName := Value;
  if (Trim(Value) <> '') then begin
    if ExtractFilePath(Value) = '' then
      case FIniLocation of
        flProgram: Value := ExtractFilePath(GetModulePath(0)) + Value;
        flModule:  Value := ExtractFilePath(GetModulePath(hInstance)) + Value;
      end;
    FIniFile := TIniFile.Create(Value);
  end;
end;

procedure TrError.SetIniLocation(Value: TrFileLocation);
begin
  if FIniLocation <> Value then begin
    FIniLocation := Value;
    if Assigned(FIniFile) then SetIniFile(FIniFileName);
  end;
end;

procedure TrError.SetMessage(Value: string);
begin
  FMessage := Value;
end;

procedure TrError.SetDescription(Value: TStrings);
begin
  FDescription.Assign(Value);
end;

function TrError.ConvertMessage(mes: string): string;
var
  j,k,l: Integer;
  b: Boolean;
  Params: TStrings;
  s: string;

  function ConvertMes(Source: string): string;
  var
    i1, i2, j1, j2: Integer;
    s1, s2, submes: string;
  begin
    Result := ConvertCodes(Source);
    // processing message including SMESSAGE
    i1 := Pos(srEHMessage, AnsiUpperCase(Result));
    if i1 <> 0 then begin
      s1 := Copy(Result, 1, i1-1);
      s2 := Copy(Result, i1 + Length(srEHMessage), Length(Result));
      submes := mes;
      repeat
        if i1 <> 1 then j1 := Pos(s1, submes) else j1 := 1;
        j2 := Length(submes) + 1;
        if j1 <> 0 then begin
          submes := Copy(submes, j1 + i1 - 1, Length(submes));
          if s2 <> '' then j2 := Pos(s2, submes);
          if j2 <> 0 then break;
        end else break;
      until False;
      Result := Copy(submes, 1, j2 - 1);
    end;
    // insert constants
    i1 := 1;
    j2 := 0;
    repeat
      j1 := Pos('<',AnsiUpperCase(Copy(Result,i1,Length(Result))));
      if j1 <> 0 then begin
        j2 := Pos('>',AnsiUpperCase(Copy(Result,i1,Length(Result))));
        if j2 <> 0 then begin
          i2 := FConsts.IndexOf(Copy(Result,i1+j1,j2-j1-1));
          if i2 <> -1 then begin
            Result := Stuff(Result,i1+j1-1,j2-j1+1,
              ConvertMes(FIniFile.ReadString(srEHConsts,FConsts.Strings[i2],'<'+FConsts.Strings[i2]+'>')));
            i1 := 1;
          end else i1 := j1+1;
        end;
      end;
    until (j1 = 0) or (j2 = 0);
  end;

begin
  Result := '';
  with FErrorBegins do begin
    for j:=0 to Count-1 do begin      { loop by ini-sections }
      s := ConvertCodes(Strings[j]);
      b := False;
      for l := 1 to WordCount(s, [';']) do begin { loop by words in ini-section }
        b := AnsiCompareText(ExtractWord(l, s, [';']),
                             Copy(mes, 1, Length(ExtractWord(l, s, [';'])))) = 0;
        if b then break;
      end;

      if b then begin
        Params := TStringList.Create;
        try
          FIniFile.ReadSection(Strings[j],Params);
          for k:=0 to Params.Count-1 do begin  { loop by keys in section }
            b := False;
            for l:=1 to WordCount(Params.Strings[k],[';']) do begin { loop by words in key }
              b := Pos(AnsiUpperCase(ConvertCodes(ExtractWord(l,Params.Strings[k],[';']))),
                       AnsiUpperCase(mes))<>0;
              if b then break;
            end;
            if b then begin
              Result := ConvertMes(FIniFile.ReadString(Strings[j],Params.Strings[k],mes));
              FSaveLog := FIniFile.ReadBool(Strings[j],srEHLogging,FSaveLog);
              break;
            end;
          end;
          if Result = '' then begin
            Result := ConvertMes(FIniFile.ReadString(Strings[j],srEHDefault,mes));
            FSaveLog := FIniFile.ReadBool(Strings[j],srEHLogging,FSaveLog);
            break;
          end;
        finally
          Params.Free;
        end;
      end;
    end;
  end;
end;

procedure TrError.AnalyzeException(E: Exception);
var
  i: Integer;
begin
  FMessage := '';
  FDescription.Clear;
  FSaveLog := False;

  if Assigned(FIniFile) then begin
    FIniFile.ReadSections(FErrorBegins);
    i := FErrorBegins.IndexOf(srEHConsts);
    if i <> -1 then begin
      FErrorBegins.Delete(i);
      FIniFile.ReadSection(srEHConsts,FConsts);
    end else FConsts.Clear;
  end;

  for i:=0 to RErrors.Count-1 do
    if AnsiCompareText(TrpEH(RErrors.Items[i])^.ErrorName, E.ClassName) = 0 then begin
      TrpEH(RErrors.Items[i])^.Proc(E, Self);
      exit;
    end;

  // unknown error
  FPreMessage := srUnkErrorTitle;
  FMessage := Format(ConvertCodes(srUnkErrorMes),[E.Message]);
  FDescription.Text := Format(srUnkErrorDesc,[E.ClassName,E.Message]);
  FSaveLog := True;
end;

procedure TrError.AppException(Sender: TObject; E: Exception);
begin
  AnalyzeException(E);
  if FSaveLog then SavingLog('');
end;

procedure TrError.SavingLog(BeforeMessage: string);
var
  F: Text;
  LogName: TFileName;
begin
  if Trim(FLogFile) = '' then exit;
  if ExtractFilePath(FLogFile) = '' then
    case FLogLocation of
      flProgram: LogName := ExtractFilePath(GetModulePath(0)) + FLogFile;
      flModule:  LogName := ExtractFilePath(GetModulePath(hInstance)) + FLogFile
    else
      LogName := FLogFile;
    end
  else
    LogName := FLogFile;
  AssignFile(F, LogName);
  try
    if FileExists(LogName) then Append(F) else Rewrite(F);
    try
      WriteLn(F, DateTimeToStr(Now), ' ',
              Format(srEHUser, [IIF(FUserName = '', GetOSUser, FUserName), GetCompName]));
      WriteLn(F, Format(srEHProgram, [ParamStr(0), GetFileVersion('')]));
      if HInstance <> MainInstance then
        WriteLn(F, Format(srEHModule, [GetModulePath(hInstance), GetFileVersion(GetModulePath(hInstance))]));
      if Length(BeforeMessage) > 0 then
        WriteLn(F, BeforeMessage);
      if FDescription.Text <> '' then WriteLn(F, FDescription.Text)
      else WriteLn(F, FMessage);
      Flush(F);
    finally
      CloseFile(F);
    end;
  except
    on E: Exception do
      FMessage := FMessage + sLineBreak + Format(ConvertCodes(srLogError), [E.Message]);
  end;
end;

function TrError.GetConstant(const aName: string): string;
begin
  if FConsts.IndexOf(aName) >= 0 then
    Result := FIniFile.ReadString(srEHConsts, aName, '')
end;

procedure DisposeErrors;
var
  i: Integer;
begin
  for i := 0 to RErrors.Count - 1 do Dispose(RErrors.Items[i]);
end;

initialization
  RErrors := TList.Create;
  AddError('EAnyError',AnyErrorEHProc);
{$IFDEF POLARIS_D6}
  AddError('EOSError',OSErrorEHProc);
{$ELSE}
  AddError('EWin32Error',OSErrorEHProc);
{$ENDIF}
  AddError('EInOutError', InOutErrorEHProc);
  AddError('EConvertError',ConvertErrorEHProc);
  AddError('ERangeError',ConvertErrorEHProc);
finalization
  DisposeErrors;
  RErrors.Free;
end.
